"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import unittest

import yahoo_panoptes.polling.polling_plugin_agent


class TestPollingPluginAgent(unittest.TestCase):
    def test_polling_plugin_task(self):
        self.assertTrue(True)
