"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This package contains modules which implement a reusable plugin based framework for the Panoptes system
"""
