"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

yahoo_panoptes.utilities namespace package
"""
__import__('pkg_resources').declare_namespace(__name__)
